@@ -14,6 +14,7 @@
  * TODO: detect [number] at compilation, optimize accordingly
  */
 
+#define IN_LIBXSLT
 #include "libxslt.h"
 
 #include <string.h>
@@ -409,6 +410,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    return(0);
 		if (step->value == NULL)
 		    continue;
+		if (step->value[0] != node->name[0])
+		    return(0);
 		if (!xmlStrEqual(step->value, node->name))
 		    return(0);
 
@@ -439,6 +442,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		if (step->value != NULL) {
 		    while (lst != NULL) {
 			if ((lst->type == XML_ELEMENT_NODE) &&
+			    (step->value[0] == lst->name[0]) &&
 			    (xmlStrEqual(step->value, lst->name)))
 			    break;
 			lst = lst->next;
@@ -451,18 +455,17 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
             case XSLT_OP_ATTR:
 		if (node->type != XML_ATTRIBUTE_NODE)
 		    return(0);
-		if (step->value == NULL)
-		    continue;
-		if (!xmlStrEqual(step->value, node->name))
-		    return(0);
-
+		if (step->value != NULL) {
+		    if (step->value[0] != node->name[0])
+			return(0);
+		    if (!xmlStrEqual(step->value, node->name))
+			return(0);
+		}
 		/* Namespace test */
 		if (node->ns == NULL) {
 		    if (step->value2 != NULL)
 			return(0);
-		} else if (node->ns->href != NULL) {
-		    if (step->value2 == NULL)
-			return(0);
+		} else if (step->value2 != NULL) {
 		    if (!xmlStrEqual(step->value2, node->ns->href))
 			return(0);
 		}
@@ -480,6 +483,8 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    return(0);
 		if (step->value == NULL)
 		    continue;
+		if (step->value[0] != node->name[0])
+		    return(0);
 		if (!xmlStrEqual(step->value, node->name))
 		    return(0);
 		/* Namespace test */
@@ -518,7 +523,9 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		while (node != NULL) {
 		    if (node == NULL)
 			return(0);
-		    if (xmlStrEqual(step->value, node->name)) {
+		    if ((node->type == XML_ELEMENT_NODE) &&
+			(step->value[0] == node->name[0]) &&
+			(xmlStrEqual(step->value, node->name))) {
 			/* Namespace test */
 			if (node->ns == NULL) {
 			    if (step->value2 == NULL)
@@ -589,18 +596,23 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		 * if the node is in the result list.
 		 */
 		if (comp->steps[i + 1].op == XSLT_OP_PREDICATE) {
-		    xmlNodePtr previous;
+		    xmlDocPtr prevdoc, doc;
 		    xmlXPathObjectPtr list;
 		    int index, j;
+		    int nocache = 0;
 
-		    previous = (xmlNodePtr)
+		    prevdoc = (xmlDocPtr)
 			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
 		    index = (int)
 			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra);
 		    list = (xmlXPathObjectPtr)
 			XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
-		    if (list == NULL) {
+		    
+		    doc = node->doc;
+		    if ((list == NULL) || (prevdoc != doc)) {
 			xmlChar *query;
+			xmlXPathObjectPtr newlist;
+			xmlNodePtr parent = node->parent;
 
 			if (comp->pattern[0] == '/')
 			    query = xmlStrdup(comp->pattern);
@@ -608,30 +620,54 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    query = xmlStrdup((const xmlChar *)"//");
 			    query = xmlStrcat(query, comp->pattern);
 			}
-			list = xmlXPathEval(query, ctxt->xpathCtxt);
+			newlist = xmlXPathEval(query, ctxt->xpathCtxt);
 			xmlFree(query);
-			if (list == NULL)
+			if (newlist == NULL)
 			    return(-1);
-			if (list->type != XPATH_NODESET) {
-			    xmlXPathFreeObject(list);
+			if (newlist->type != XPATH_NODESET) {
+			    xmlXPathFreeObject(newlist);
 			    return(-1);
 			}
-			XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
-			    (void *) list;
-			XSLT_RUNTIME_EXTRA_FREE(ctxt, select->lenExtra) =
-			    (xmlFreeFunc) xmlXPathFreeObject;
+			
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
+			if (nocache == 0) {
+			    if (list != NULL)
+				xmlXPathFreeObject(list);
+			    list = newlist;
+
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) list;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				(void *) doc;
+			    XSLT_RUNTIME_EXTRA_FREE(ctxt, select->lenExtra) =
+				(xmlFreeFunc) xmlXPathFreeObject;
+			}
 		    }
 		    if ((list->nodesetval == NULL) ||
 			(list->nodesetval->nodeNr <= 0))
 			return(0);
+		    /* TODO: store the index and use it for the scan */
 		    if (index == 0) {
 			for (j = 0;j < list->nodesetval->nodeNr;j++) {
 			    if (list->nodesetval->nodeTab[j] == node) {
+				if (nocache == 1)
+				    xmlXPathFreeObject(list);
 				return(1);
 			    }
 			}
 		    } else {
 		    }
+		    if (nocache == 1)
+			xmlXPathFreeObject(list);
 		    return(0);
 		}
 		/*
@@ -648,7 +684,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    (node->type == XML_ELEMENT_NODE) &&
 		    (node->parent != NULL)) {
 		    xmlNodePtr previous;
-		    int index;
+		    int index, nocache = 0;
 
 		    previous = (xmlNodePtr)
 			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
@@ -665,7 +701,9 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			while (sibling != NULL) {
 			    if (sibling == previous)
 				break;
-			    if (xmlStrEqual(node->name, sibling->name)) {
+			    if ((node->type == XML_ELEMENT_NODE) &&
+				(node->name[0] == sibling->name[0]) &&
+				(xmlStrEqual(node->name, sibling->name))) {
 				if ((select->value2 == NULL) ||
 				    ((sibling->ns != NULL) &&
 				     (xmlStrEqual(select->value2,
@@ -691,12 +729,18 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			}
 			if (sibling != NULL) {
 			    pos = index + indx;
-			    len = (int)
-				XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
-			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
-			        node;
-			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
-			        (void *) pos;
+			    /*
+			     * If the node is in a Value Tree we cannot
+			     * cache it !
+			     */
+			    if (node->doc != NULL) {
+				len = (int)
+				    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
+				XSLT_RUNTIME_EXTRA(ctxt,
+					select->previousExtra) = node;
+				XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				    (void *) pos;
+			    }
 			    index = pos;
 			} else
 			    pos = 0;
@@ -705,14 +749,15 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * recompute the index
 			 */
 			xmlNodePtr siblings = node->parent->children;
+			xmlNodePtr parent = node->parent;
 
 			while (siblings != NULL) {
 			    if (siblings->type == XML_ELEMENT_NODE) {
 				if (siblings == node) {
 				    len++;
 				    pos = len;
-				} else if (xmlStrEqual(node->name,
-					   siblings->name)) {
+				} else if ((node->name[0] == siblings->name[0])
+			       && (xmlStrEqual(node->name, siblings->name))) {
 				    if ((select->value2 == NULL) ||
 					((siblings->ns != NULL) &&
 					 (xmlStrEqual(select->value2,
@@ -722,21 +767,37 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			    siblings = siblings->next;
 			}
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
 		    }
 		    if (pos != 0) {
 			ctxt->xpathCtxt->contextSize = len;
 			ctxt->xpathCtxt->proximityPosition = pos;
-			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
-			    node;
-			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
-			    (void *) pos;
-			XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
-			    (void *) len;
+			/*
+			 * If the node is in a Value Tree we cannot
+			 * cache it !
+			 */
+			if ((node->doc != NULL) && (nocache == 0)) {
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				node;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				(void *) pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) len;
+			}
 		    }
 		} else if ((select != NULL) && (select->op == XSLT_OP_ALL) &&
 			   (node->type == XML_ELEMENT_NODE)) {
 		    xmlNodePtr previous;
-		    int index;
+		    int index, nocache = 0;
 
 		    previous = (xmlNodePtr)
 			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra);
@@ -771,12 +832,18 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			}
 			if (sibling != NULL) {
 			    pos = index + indx;
-			    len = (int)
-				XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
-			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
-			        node;
-			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
-			        (void *) pos;
+			    /*
+			     * If the node is in a Value Tree we cannot
+			     * cache it !
+			     */
+			    if (node->doc != NULL) {
+				len = (int)
+				    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra);
+				XSLT_RUNTIME_EXTRA(ctxt,
+					select->previousExtra) = node;
+				XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				    (void *) pos;
+			    }
 			} else
 			    pos = 0;
 		    } else {
@@ -784,6 +851,7 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			 * recompute the index
 			 */
 			xmlNodePtr siblings = node->parent->children;
+			xmlNodePtr parent = node->parent;
 
 			while (siblings != NULL) {
 			    if (siblings->type == XML_ELEMENT_NODE) {
@@ -794,16 +862,32 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 			    }
 			    siblings = siblings->next;
 			}
+			if ((parent == NULL) || (node->doc == NULL))
+			    nocache = 1;
+			else {
+			    while (parent->parent != NULL)
+				parent = parent->parent;
+			    if (((parent->type != XML_DOCUMENT_NODE) &&
+				 (parent->type != XML_HTML_DOCUMENT_NODE)) ||
+				 (parent != (xmlNodePtr) node->doc))
+				nocache = 1;
+			}
 		    }
 		    if (pos != 0) {
 			ctxt->xpathCtxt->contextSize = len;
 			ctxt->xpathCtxt->proximityPosition = pos;
-			XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
-			    node;
-			XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
-			    (void *) pos;
-			XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
-			    (void *) len;
+			/*
+			 * If the node is in a Value Tree we cannot
+			 * cache it !
+			 */
+			if ((node->doc != NULL) && (nocache == 0)) {
+			    XSLT_RUNTIME_EXTRA(ctxt, select->previousExtra) =
+				node;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->indexExtra) =
+				(void *) pos;
+			    XSLT_RUNTIME_EXTRA(ctxt, select->lenExtra) =
+				(void *) len;
+			}
 		    }
 		}
 		oldNode = ctxt->node;
@@ -1300,7 +1384,7 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	    goto error;
 	}
 	PUSH(XSLT_OP_ATTR, token, URL);
-	return;
+	goto parse_predicate;
     }
     if (token == NULL)
 	token = xsltScanName(ctxt);
@@ -1690,8 +1774,14 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc,
 			 element->pattern);
 #endif
 	xsltCompileLocationPathPattern(ctxt);
-	if (ctxt->error)
+	if (ctxt->error) {
+	    xsltPrintErrorContext(NULL, style, node);
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltCompilePattern : failed to compile '%s'\n",
+			     element->pattern);
+	    style->errors++;
 	    goto error;
+	}
 
 	/*
 	 * Reverse for faster interpretation.
@@ -1751,6 +1841,7 @@ xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc,
 	xsltPrintErrorContext(NULL, NULL, node); /* TODO */
         xsltGenericError(xsltGenericErrorContext,
 			 "xsltCompilePattern : NULL pattern\n");
+	style->errors++;
 	goto error;
     }
 
@@ -1817,7 +1908,6 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	    else
 		top = (xsltCompMatchPtr *) &(style->attrMatch);
 	    break;
-        case XSLT_OP_ELEM:
         case XSLT_OP_CHILD:
         case XSLT_OP_PARENT:
         case XSLT_OP_ANCESTOR:
@@ -1858,12 +1948,12 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	case XSLT_OP_TEXT:
 	    top = (xsltCompMatchPtr *) &(style->textMatch);
 	    break;
+        case XSLT_OP_ELEM:
 	case XSLT_OP_NODE:
 	    if (pat->steps[0].value != NULL)
 		name = pat->steps[0].value;
 	    else
 		top = (xsltCompMatchPtr *) &(style->elemMatch);
-	    
 	    break;
 	}
 	if (name != NULL) {
